/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 * Purpose: Injects methods into the interface IAccessClassified to enable
 * 			objects of its implementing classes in classifying the access-type
 * 			(READ, WRITE, UPDATE)of each of its method.
 */
package ca.mcgill.acid.aspects;

import java.lang.reflect.*;
import ca.mcgill.acid.utility.Read;
import ca.mcgill.acid.utility.Write;
import ca.mcgill.acid.utility.Update;
import ca.mcgill.acid.utility.MethodNotAnnotatedException;
import ca.mcgill.acid.utility.MethodNotFoundException;
import ca.mcgill.acid.interfaces.Constants;
import ca.mcgill.acid.interfaces.IAccessClassified;

public aspect AccessClassified { 
	
	private boolean IAccessClassified.found;
	private boolean IAccessClassified.implementsIShared = false;
	private boolean IAccessClassified.annotated;
	private String IAccessClassified.accessType = null;
	
	/**
	 * @param String
	 * @return String
	 * Purpose: Classifiers each method(Declared, Inherited, Injected) 
	 * of the invoking object as either READ, WRITE or UPDATE. 		
	 */
	public String IAccessClassified.getAccessType(String methodName)
		throws MethodNotAnnotatedException,MethodNotFoundException
	{ 
		found = false;
		annotated = false;
		
		for (Method m : this.getClass().getMethods()) 
		{
			if((methodName.trim()).equalsIgnoreCase(m.getName()))
			{
		         if (m.isAnnotationPresent(Read.class) 
		          || m.isAnnotationPresent(Write.class)
		          || m.isAnnotationPresent(Update.class)) {
		        	 
		        	 if (m.isAnnotationPresent(Read.class))
		        		 accessType = Constants.READ;

		        	 if (m.isAnnotationPresent(Write.class))
		        		 accessType = Constants.WRITE;

		        	 if (m.isAnnotationPresent(Update.class))
		        		 accessType = Constants.UPDATE;
		            
		        	 annotated = true;
		         }	
		         
		       found = true;  
		       break;
		   }
	   }
		
		if(!found)
       	 	throw new MethodNotFoundException(
       	 	"\n	The Method <" + methodName + 
       	 	"> is not defined in the class \n" + 
       	 	"	<" + this.getClass().getName() + "> or \n" + 
       	 	"	its super classes ot its Interfaces\n");

		if(found && !annotated)
		{
			 accessType = Constants.UPDATE;
		}
		
		return accessType;
	}
	
	/**
	 * Helper method for the one above
	 *
	private String IAccessClassified.classsifyAccessType(Class _class, String _name)
	throws MethodNotAnnotatedException,MethodNotFoundException
	{
		String type = null;
		annotated = false;
		found = false;
		
		for (Method m : _class.getMethods()) 
		{
			if((_name.trim()).equalsIgnoreCase(m.getName()))
			{
		         if (m.isAnnotationPresent(Read.class) 
		          || m.isAnnotationPresent(Write.class)
		          || m.isAnnotationPresent(Update.class)) {
		        	 
		        	 if (m.isAnnotationPresent(Read.class))
		        		 type = Constants.READ;

		        	 if (m.isAnnotationPresent(Write.class))
		        		 type = Constants.WRITE;

		        	 if (m.isAnnotationPresent(Update.class))
		        		 type = Constants.UPDATE;
		            
		        	 annotated = true;
		         }	
		         
		       found = true;  
		       break;
		   }
	   }
			
		return type;
	}
	
	
	private void IAccessClassified.implementsInterface(Class className)
	 {     	
		   for(Class _class:className.getInterfaces())
		   {
			   if(_class.getName().endsWith(Constants.ISHARED))
	  			{
				   accessType = Constants.UPDATE;
				   implementsIShared = true;
	  			   break;
	  			}
	  			else
	  				this.implementsInterface(_class);
		   }   
	 }

	
	accessType = classsifyAccessType(obj.getClass(), methodName);
	
	if(found && !annotated)
	{ 
		superClass = obj.getClass();
		
		   while(superClass != null)
		   {
		   		for(Class _class:superClass.getInterfaces())
		   		{
		   			if(_class.getName().endsWith(Constants.ISHARED))
		   			{
		   				accessType = Constants.UPDATE;
		   				implementsIShared = true;
		   				break;
		   			}
		   			else
		   				this.implementsInterface(_class);
		   		}
		   		
		   		//	Check all super classes
		   		superClass = superClass.getSuperclass();
		   }
	}
	
	if(!found)
   	 	throw new MethodNotFoundException(
   	 	"\n	The Method <" + methodName + 
   	 	"> is not defined in the class \n" + 
   	 	"	<" + obj.getClass().getName() + "> or \n" + 
   	 	"	its super classes ot its Interfaces\n");

	if( found && !annotated && !implementsIShared)
		throw new MethodNotAnnotatedException(
		"\n	The Method <" + methodName  + "> invoked from Class " + 
		"<" + obj.getClass().getName() + 
		"> \n	must be Annotated as either:"
		+ " @Read, @Write or @Update at definition point\n");
	*/
}


