 /**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 */
package ca.mcgill.acid.aspects;

import ca.mcgill.acid.interfaces.ICopyable;
import ca.mcgill.acid.interfaces.Constants;
import ca.mcgill.acid.interfaces.IShared;
import java.lang.reflect.Field;
import java.lang.reflect.AccessibleObject; 
import java.lang.reflect.Modifier;
import ca.mcgill.acid.utility.SourceClassNotEqualDestinationClass;
import ca.mcgill.acid.utility.Lock;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;

 
public aspect Copyable { 
	
	declare parents: ICopyable implements Cloneable; //Enables cloning of classes that implements ICopyable

/**
 * Introduce a replaceState(source) method to every class that implements
 * the ICopyable interface. It enables an object to replace its state with
 * that of another object of the same Class
 */	 
	public void ICopyable.replaceState( Object source ) { 
		 
		try{ 
			copyFields(this,source);
		}catch(SourceClassNotEqualDestinationClass e){
			e.printStackTrace();
		} 
	}
	
	/**
	 * Introduce a clone() method to every class that implements
	 * the ICopyable interface. This method provides a deep copy
	 * of a serialized object.
	 */
	public Object ICopyable.clone(){ 
		
		Object deepCopyOfOriginalObject = null;
		
		try{
		   ByteArrayOutputStream memoryOutputStream = new ByteArrayOutputStream( );
		   ObjectOutputStream serializer = new ObjectOutputStream(memoryOutputStream);
		   serializer.writeObject(this);
		   serializer.flush();
		    
		   ByteArrayInputStream memoryInputStream = new ByteArrayInputStream(memoryOutputStream.toByteArray());
		   ObjectInputStream deserializer = new ObjectInputStream(memoryInputStream);
		   deepCopyOfOriginalObject = deserializer.readObject();
		}catch (IOException e ){
	          	e.printStackTrace();
        }catch (ClassNotFoundException e ){
	          	e.printStackTrace();
        }
	        return deepCopyOfOriginalObject;
	}   
	 
	/**
	 * Helper method for ICopyable.replaceState( Object source )
	 */
	private static void copyFields(Object destination, Object source)
	throws SourceClassNotEqualDestinationClass {
 
		if(!(destination.getClass()== source.getClass()))
			throw new SourceClassNotEqualDestinationClass(
					"\n	<< Name of soure Class not equals to Destination Class >> \n");
		 
		Class targetClass = source.getClass();
		
		// Recursively copy the values from "source" to "this";
		// including all variables inherited from superclasses,
		// interfaces, inter-type declaration and composed objects
		while(targetClass != null)
		{

		 Field[] thisFields  = targetClass.getDeclaredFields();
         Field[] otherFields = targetClass.getDeclaredFields();

         // make the protected and private fields accessible
         AccessibleObject.setAccessible(thisFields, true);
         AccessibleObject.setAccessible(otherFields, true);

         // perform field by field copy of non-final variables
         for (int i = 0; i < thisFields.length; i++) 
         {
        	 if(!Modifier.isFinal(thisFields[i].getModifiers()))
        	 {
        		 try 
        		 {
        			 //Do not over-write an Object's Name because it's its identity. 
        			 //This is the Named field introduced by the aspect Named.aj.
        			 if ((thisFields[i].getType().isPrimitive() || thisFields[i].getType().equals(String.class)
        				||otherFields[i].get(source) == null) && !thisFields[i].getName().endsWith(Constants.OBJECT_NAME))
        			 {
        				 thisFields[i].set(destination, otherFields[i].get(source));
        			 }
        			 else
        			 {
        				 // Do not over-write the Lock field of an Object
        				 if(!thisFields[i].getType().equals(Lock.class))
        					 copyFields(thisFields[i].get(destination),otherFields[i].get(source));
        			 }
        	 	 } catch (IllegalAccessException e) {
        	 			
        	 			e.printStackTrace();
       				 	
        	 	}
        	 }
          }
		
         // Recursively copy the values of all super classes
         targetClass = targetClass.getSuperclass();
		
		}
	}
    
}
