/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 *
 * Purpose: Assigns a name to the objects of classes that implements the interface
 * 			INamed. 
 */

package ca.mcgill.acid.aspects;

import ca.mcgill.acid.interfaces.INamed;
import java.util.Vector;
import ca.mcgill.acid.utility.ObjectNotNamedException;

public aspect Named {
	
	private static int ID = 0;
	private Vector<INamed> namedObjects = new Vector<INamed>();
	
	private String INamed.myName; 
	public String INamed.getMyName()  { return this.myName; }
    private void INamed.setMyName(String _name)  { myName = _name; }
     
    /**
     * Constructor execution on the target object INamed
     */
    pointcut constructorExecution(INamed iname):
    	target(iname) && execution(public INamed+.new(..));
    
    /**
     *  Assigns a unique name to every object of type INamed after
     *  its creation
     */
    after(INamed iname) returning: constructorExecution(iname)
    {
    	++ID;
    	iname.setMyName(iname.getClass().getName() + ID);
    	namedObjects.add(iname);
    }

    /**
     * @param String
     * @return INamed
     * Purpose: returns an INamed object of the given name
     */
    public  synchronized INamed getObject(String name) throws ObjectNotNamedException
    {
    	INamed obj = null;
    	INamed tempObj = null;
    	boolean found = false;
    	
    	for(int i = 0; i < namedObjects.size(); i++)
    	{
    		tempObj = namedObjects.elementAt(i);

    		if((name.trim()).equalsIgnoreCase(tempObj.getMyName()))
    		{
    			found = true;
    			obj = tempObj;
    			break;
    		}		
    	} 
    		 
    	if(!found)
    		throw new ObjectNotNamedException(
    				"\n	Object not Named at creation time\n");
    	
    	return obj;
    }
    
}
    


    
   


