/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 * Purpose: Enables objects that implements the IRecoverable interface
 * 			to save and restore their state.
 */
package ca.mcgill.acid.aspects;

import ca.mcgill.acid.interfaces.IVersioned;
import ca.mcgill.acid.interfaces.IShared;
import ca.mcgill.acid.utility.Version;
import ca.mcgill.acid.interfaces.IRecoverable;
import ca.mcgill.acid.utility.View;
import ca.mcgill.acid.utility.ObjectNotFoundException;

public aspect Recoverable {

	declare parents: IRecoverable implements IVersioned,IShared;
 
	private boolean IRecoverable.deferred = false; 

	/**
	 * Create a new checkpoint for the "this" object.
	 * The View of the invoking thread will continue to
	 * point to the "this" object if the update strategy
	 * is In-Place update. Otherwise[i.e Deferred update],
	 *  the view will now point to the newly created version
	 *   of the object. 
	 */
	public int IRecoverable.establishCheckpoint(View view) 
	{
		int checkpointID= -1;
		int thisID = -1;
		boolean found = false;
		Version v = null;
		
			thisID = this.getVersionID();
			v = this.newVersion();
			checkpointID = v.getVersionID();
			
		if(deferred && view != null)
		{
			for(int i = 0; i < view.getObjectsInView().size(); i++ )
			{
			  if(((Version)view.getObjectsInView().elementAt(i)).getObjectName().equals(this.getMyName()))
			  {
				  ((Version)view.getObjectsInView().elementAt(i)).setVersionID(checkpointID);
				  found = true;
				  break;
			  }
			}
		}
		
		if(!deferred && view != null)
		{
			for(int i = 0; i < view.getObjectsInView().size(); i++ )
			{
			  if(((Version)view.getObjectsInView().elementAt(i)).getObjectName().equals(this.getMyName()))
			  {
				  ((Version)view.getObjectsInView().elementAt(i)).setVersionID(thisID);
				  found = true;
				  break;
			  }
			}
		}
				
		if(!found && view != null)
			view.addObject(v);
		
		return checkpointID;
	}
	
	/**
	 * Changes the version of the "this" object in the View to checkpointID.
	 * This enables a Recoverable object to restore its state from any of the
	 * previously saved/checkpointed states
	 */
	public void IRecoverable.restoreCheckpoint(int checkpointID, View view) 
		throws ObjectNotFoundException
	{
		boolean found = false;
		
		for(int i = 0; i < view.getObjectsInView().size(); i++ )
		{
		  if(((Version)view.getObjectsInView().elementAt(i)).getObjectName().equals(this.getMyName()))
		  {
			  ((Version)view.getObjectsInView().elementAt(i)).setVersionID(checkpointID);
			  found = true;
			  break;
		  }
		}
		
		if(!found)
			throw new ObjectNotFoundException("\n	Version<" +checkpointID+ "> of the Object <" 
												+ this.getMyName() + "> cannot be found");
	}	
	
	/**
	 * Deletes the checkpoint with ID checkpointID of the "this" object
	 */
	public void IRecoverable.discardCheckpoint(int checkpointID) 
	{
		this.deleteVersion(checkpointID); 
	}
	
	/**
	 * Enables the "this" object to change it's update strategy
	 * from In-Place to Deferred and vice-versa
	 */
	public void IRecoverable.setDeferred(boolean On) 
	{
		this.deferred = On; 
	}
}
