/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 *
 * Purpose: Synchronizes access to the objects of all classes that implements
 *  		the IShared interface. It's a reader's preference technique.
 */
package ca.mcgill.acid.aspects;

import ca.mcgill.acid.utility.Lock;
import ca.mcgill.acid.interfaces.IShared;
import ca.mcgill.acid.interfaces.IAccessClassified;
import java.util.StringTokenizer;
import ca.mcgill.acid.utility.MethodNotAnnotatedException;
import ca.mcgill.acid.utility.MethodNotFoundException;
import ca.mcgill.acid.interfaces.Constants;

public aspect Shared { 
	   
	declare parents: IShared implements IAccessClassified;

	//Introduce the variable and method for enforcing synchronization
	private Lock IShared.threadLock = new Lock();
	private Lock IShared.getLock()  { return threadLock; }
	
	//Introduce the variables and methods for runtime disabling
	//of the shared advice
	public boolean IShared.ENABLED = true;
	private boolean IShared.getEnabled() { return ENABLED; } 
	private static boolean isEnabled(IShared object) { return object.getEnabled(); }

	
	/**
     * Method Execution on the target object IShared: 
     * Captures both reflective and non-reflective executions
     * to ensure absolute synchronization
     */
	pointcut methodExecution( IShared ishare ): 
		target(ishare) && call(public * IShared+.*(..)) ;
	  

	/**
	 * Synchronizes access to the data members of IShared objects
	 * by ensuring that conflicting operations are not executed
	 * concurrently by Threads
	 */
	Object around(IShared shared) :  methodExecution(shared) && if(isEnabled(shared))
	{
		Object obj;
		String accessType = null;
		
		try{
			//Get the access-type of the invoked method
			accessType = shared.getAccessType(getMethodName(thisJoinPoint.toShortString()));		
		}
		catch (MethodNotAnnotatedException e)
		{e.printStackTrace();}
		catch (MethodNotFoundException e)
		{e.printStackTrace();}
		
		System.out.println("\n 		<- " + getMethodName(thisJoinPoint.toShortString()));
		
		// Get the appropriate lock
		if(Constants.READ.equalsIgnoreCase(accessType)){
			if(!Thread.currentThread().getName().equals(shared.getLock().getWriteLockHolder())
			&& !Thread.currentThread().getName().equals(shared.getLock().getUpdateLockHolder())
			&& !Thread.currentThread().getName().equals(shared.getLock().getReadLockHolder()))
			{ 
				shared.getLock().getReadLock(Thread.currentThread().getName(),getMethodName(thisJoinPoint.toShortString())); 
				System.out.println(" 			Get Read Lock " );
			}
		}

		if(Constants.WRITE.equalsIgnoreCase(accessType)){
			if(!Thread.currentThread().getName().equals(shared.getLock().getWriteLockHolder())
			&& !Thread.currentThread().getName().equals(shared.getLock().getUpdateLockHolder()))	
			{
				if(Thread.currentThread().getName().equals(shared.getLock().getReadLockHolder()))
					{
					 System.out.println(" 			Release Read Lock ");
					 shared.getLock().releaseReadLock();
					}
				
				shared.getLock().getWriteLock(Thread.currentThread().getName(),getMethodName(thisJoinPoint.toShortString()));
				System.out.println(" 			Get Write Lock ");
			}
		}
		 
		if(Constants.UPDATE.equalsIgnoreCase(accessType)) {

			if(!Thread.currentThread().getName().equals(shared.getLock().getWriteLockHolder())
			&& !Thread.currentThread().getName().equals(shared.getLock().getUpdateLockHolder()))					
			{	
				
				if(Thread.currentThread().getName().equals(shared.getLock().getReadLockHolder()))
				{
					 System.out.println(" 			Release Read Lock ");
					 shared.getLock().releaseReadLock();
				}
				
				shared.getLock().getUpdateLock(Thread.currentThread().getName(),getMethodName(thisJoinPoint.toShortString()));
				System.out.println(" 			Get Update lock ");
			}  
			
		} 
		
		obj = proceed(shared);
		
		// Release the appropriate lock
		if(Constants.READ.equalsIgnoreCase(accessType)){
			if(Thread.currentThread().getName().equals(shared.getLock().getReadLockHolder())
			&& getMethodName(thisJoinPoint.toShortString()).equals(shared.getLock().getMethodName()))	
			{ 
			  shared.getLock().releaseReadLock();	
			  System.out.println(" 			Release Read Lock ");
			}
		}

		if(Constants.WRITE.equalsIgnoreCase(accessType)){
			if(Thread.currentThread().getName().equals(shared.getLock().getWriteLockHolder())
			&& getMethodName(thisJoinPoint.toShortString()).equals(shared.getLock().getMethodName()))	
			{
				shared.getLock().releaseWriteLock();	
				System.out.println(" 			Release Write Lock ");
			}
		}
		
		if(Constants.UPDATE.equalsIgnoreCase(accessType)){
			
			if(Thread.currentThread().getName().equals(shared.getLock().getUpdateLockHolder())
			&& getMethodName(thisJoinPoint.toShortString()).equals(shared.getLock().getMethodName()))	
				
				{
					shared.getLock().releaseUpdateLock();
					System.out.println(" 			Release Update Lock ");

				}

		}
		
		System.out.println(" 		"+ getMethodName(thisJoinPoint.toShortString()) + " -> ");		

		return obj;
	}
	
	 
	/**
	 * Helper method to around advice
	 */
	private synchronized String getMethodName(String s)
	{

		StringTokenizer st = new StringTokenizer(s,"."); 
		String temp = st.nextToken();
		
		StringTokenizer st2 = new StringTokenizer(st.nextToken(), "(");
		
		return (st2.nextToken()).trim();
	}
} 
