/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 * Purpose: It encapsulates serveral states of an IVersioned objects. 
 * 			
 * 			
 */
package ca.mcgill.acid.aspects;

import ca.mcgill.acid.interfaces.IVersioned;
import ca.mcgill.acid.interfaces.INamed;
import ca.mcgill.acid.interfaces.IShared;
import ca.mcgill.acid.interfaces.ICopyable;
import ca.mcgill.acid.utility.View;
import ca.mcgill.acid.utility.NoViewFoundException;
import ca.mcgill.acid.utility.Version;
import java.util.Vector;
import java.util.StringTokenizer;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import ca.mcgill.acid.utility.ObjectNotFoundException;


public aspect Versioned {

	declare parents: IVersioned implements INamed, ICopyable; 

	private static InheritableThreadLocal myView = new InheritableThreadLocal();
	
	private static Vector versionedObjects = new Vector();
	
	private boolean IVersioned.mainVersion;  
	private int IVersioned.versionID = 0;
	private int IVersioned.previousVersionID = 0;
    private void IVersioned.setVersionID(int id)  { versionID = id; }
    public int IVersioned.getVersionID()  { return versionID; }
    private void IVersioned.setMainVersion(boolean v)  { mainVersion = v; }
    private boolean IVersioned.isMainVersion()  { return mainVersion; }
    
    /**
     * Constructor execution on the target object IVersioned
     */
    pointcut constructorExecution(IVersioned versioned):
    	target(versioned) && execution(public IVersioned+.new(..));
	
    /**
     * Method call on the target object IVersioned
     */
    pointcut methodCall(IVersioned versioned): 
		target(versioned) && call(public !static * IVersioned+.*(..))
		&& !call(public * IVersioned+.newVersion()) 
		&& !call(public * IVersioned+.setToMainVersion())
		&& !call(public * IVersioned+.deleteVersion())
		&& !call(public * IVersioned+.getVersionID());
	
    /**
     * An instantiated object is by default the main version. It is assigned
     * the versionID 0
     */
    after(IVersioned versioned) returning: constructorExecution(versioned)
    { 
    	versioned.setVersionID(0);
    	versioned.setMainVersion(true);
    	versionedObjects.add(versioned);
    
    }

    /**
     * Determines what version of an object to execute a method call.
     * The version of an object in the view of the invoking thread is by
     * default the target of the invocation.The call is directed to the main 
     * version if the invoking thread has no view or a version of the target
     * object 
     */
    Object around(IVersioned versioned) :  methodCall(versioned)
	{
		Object returnedObj = null;
		int objID = -1;
		String objName = versioned.getMyName();
		
		 if((myView.get()!= null) && objectInMyView(myView.get(),objName))
		 {
			 for(int i = 0; i < ((View)myView.get()).getObjectsInView().size(); i++ )
			 {
				  if(((Version)((View)myView.get()).getObjectsInView().elementAt(i)).getObjectName().equals(objName))
				  {
					  objID = ((Version)((View)myView.get()).getObjectsInView().elementAt(i)).getVersionID();
					  break;
				  }
			 }
			 //System.out.println("	 --> Myview - ID: "+objID); 
		 }
		 else
		 {
			 for(int i = 0; i < versionedObjects.size(); i++ )
			 {
			 	
			 	if(((IVersioned)versionedObjects.elementAt(i)).isMainVersion() 
					  && ((INamed)versionedObjects.elementAt(i)).getMyName().equals(objName))
				  {
					  objID = ((IVersioned)versionedObjects.elementAt(i)).getVersionID();
					  break;
				  }
				
			 }
			 //System.out.println("	 --> Main View -  ID: "+objID);
		 }
		
		 //Invoke method on the appropriate version of the target object
		 try{
			 returnedObj = invokeMethod(objName,objID,getMethodName(thisJoinPoint.toShortString()),thisJoinPoint.getArgs());
		 }catch(ObjectNotFoundException e)
		 {
			 e.printStackTrace();
		 }
		 
		return returnedObj;
	}
    

    /**
     * Create a new version of the "this" object for the invoking thread and
     * assign it to its view; if it exist. Each new version is assigned a 
     * unique versionID and is by default no the main version
     */
	public Version IVersioned.newVersion(){ 
		Version v = null;
		
		if((myView.get()!= null) && !objectInMyView(myView.get(),this.getMyName()))
		 {
			++previousVersionID;
			IVersioned versionObj = (IVersioned)this.clone();
			versionObj.setVersionID(previousVersionID);
			versionObj.setMainVersion(false);
			v = new Version(previousVersionID,versionObj.getMyName());
			((View)myView.get()).addObject(v);	
			versionedObjects.add(versionObj);
			
		 }
		else if((objectInMyView(myView.get(),this.getMyName())) || (myView.get() == null)) 
		{
			++previousVersionID;
			IVersioned versionObj = (IVersioned)this.clone();
			versionObj.setVersionID(previousVersionID);
			versionObj.setMainVersion(false);
			versionedObjects.add(versionObj);
			v = new Version(previousVersionID,versionObj.getMyName());
				
		}
		
		return v;
	} 
	
	/**
	 * Deletes the version with identity ID of the "this" object
	 */
	public void IVersioned.deleteVersion(int ID) { 
		
		 for(int i = 0; i < versionedObjects.size(); i++)
			 if(((IVersioned)versionedObjects.elementAt(i)).getMyName().equals(this.getMyName())
			 && (((IVersioned)versionedObjects.elementAt(i)).getVersionID() == ID))
				 versionedObjects.remove(i);	 
	} 
	
	/**
	 * Sets the "this" version to be the MainVersion
	 */
	public void IVersioned.setToMainVersion() { 
			
		 for(int i = 0; i < versionedObjects.size(); i++)
			 if(((IVersioned)versionedObjects.elementAt(i)).getMyName().equals(this.getMyName()))
				 ((IVersioned)versionedObjects.elementAt(i)).setMainVersion(false);
		 
		 this.setMainVersion(true);
	} 
	
	/**
	 * Creat a new view for the invoking thread
	 */
	public static synchronized View newView()
	{ 
		if(myView.get()== null)
		{
			View view = new View();
			myView.set(view);
		}
		
		return (View)myView.get();
	}
	
	/**
	 * Join an existing view
	 */
	public static synchronized void joinView(View view)
	{
		 myView.set(view);
	}
	
	/**
	 * Delete the view of the invoking thread
	 */
	public static synchronized void deleteView()
	{
		 myView.set(null);
	}
	
	/**
	 * Returns the current version of the object with Name objectName 
	 * in the view of the Thread
	 */
	public static synchronized Version getCurrentVersion(String objectName) 
		throws NoViewFoundException
	{
		if(myView.get() == null)
			throw new NoViewFoundException("\n	The Thread <" + Thread.currentThread().getName()+
					"> must have a View before invoking the method Version.getCurrentVersion(..)");
        
		Vector v = null;
		Version version = null;
		
		v = ((View)myView.get()).getObjectsInView();
		
		for(int i = 0; i < v.size(); i++)	
		{
			if(objectName.trim().equalsIgnoreCase(((Version)v.elementAt(i)).getObjectName()))
				{
					version = (Version)v.elementAt(i);
					break;
				}
		}

		if(version == null)
			throw new NoViewFoundException("\n	The Thread <" + Thread.currentThread().getName()+
			"> has no version of the Object <" +objectName+ "> in its View");

	
		return version;
	}
	
	
	/**
	 * Set the current version of the object with Name objectName in the view 
	 * of the invoking thread to versionID
	 */
	public static synchronized void setCurrentVersion(String objectName, int versionID) 
		throws NoViewFoundException
	{
		if(myView.get() == null)
			throw new NoViewFoundException("\n	The Thread <" + Thread.currentThread().getName()+
					"> must have a View before invoking the method Versioned.setCurrentVersion(..)");
		
		Vector v = null;
		Version version = null;
		v = ((View)myView.get()).getObjectsInView();
		
		for(int i = 0; i < v.size(); i++)	
		{
			if(objectName.trim().equalsIgnoreCase(((Version)v.elementAt(i)).getObjectName()))
				{
					version = (Version)v.elementAt(i);
					break;
				}
		}
		
		if(version == null)
		{
			Version vID = new Version(versionID,objectName);
			((View)myView.get()).addObject(vID);
		}
		else
			version.setVersionID(versionID);
	}
	
	
	/**
	 * Return all versioned objects
	 */
	public static synchronized Vector getObjects() 
	{
		return versionedObjects;
	}
	
	/**
	 * Helper method
	 */
	private static synchronized boolean objectInMyView(Object obj, String objName)
	{
		boolean found = false;
		
		if(obj == null)
			return found;
		
		for(int i = 0; i < ((View)obj).getObjectsInView().size(); i++ )
		{
		  if(((Version)((View)obj).getObjectsInView().elementAt(i)).getObjectName().equals(objName))
		  {
			  found = true;
			  break;
		  }
		}
		
		return found;
	}
	
	/**
	 * Helper method
	 */
	private synchronized Object invokeMethod(String objectName, int objID, String methodName, Object[] parameters)
	throws ObjectNotFoundException
	{
		Object targetObject = null, returnedObject = null;
		boolean found = false;
		
		for(int i = 0; i < versionedObjects.size(); i++ )
		 {
			//System.out.println("	Obj: " + ((IVersioned)versionedObjects.elementAt(i)).getMyName()
								//+ "	ID: " + ((IVersioned)versionedObjects.elementAt(i)).getVersionID());
			  if(((IVersioned)versionedObjects.elementAt(i)).getVersionID() == objID
				  && ((IVersioned)versionedObjects.elementAt(i)).getMyName().equals(objectName))
			  {
				  targetObject = versionedObjects.elementAt(i);
				  found = true;
				  break;
			  }
		 }
		
		if(!found)
			throw new ObjectNotFoundException("\n	Version<" +objID+ "> of the Object <" 
												+ objectName + "> cannot be found\n");
	
		System.out.println("		"+methodName+" @ Version<" +
				(objID == 0 ? "MAIN": "" + objID ) + "> For_Object <" 
				+ objectName + ">");
		 
		for(Method m: targetObject.getClass().getMethods())
		{
			if(m.getName().equals(methodName))
			{
				try{
					returnedObject = m.invoke(targetObject,parameters);
				}catch(IllegalAccessException e){
					e.printStackTrace();
				}catch(IllegalArgumentException e){
					e.printStackTrace();
				}catch(InvocationTargetException e){
					e.printStackTrace();
				}
				break;
			}
		}
		
		return returnedObject;
	}
	
	/**
	 * Helper method
	 */
	private synchronized String getMethodName(String s)
	{

		StringTokenizer st = new StringTokenizer(s,"."); 
		String temp = st.nextToken();
		
		StringTokenizer st2 = new StringTokenizer(st.nextToken(), "(");
		
		return (st2.nextToken()).trim();
	}
}
