package ca.mcgill.acid.utility;

/**
 * Provides a reader's preference synchronization
 * scheme to a shared resource. 
 * 
 */


import ca.mcgill.acid.utility.Semaphore;

import java.io.Serializable;

public class Lock implements Serializable{
	
	transient private int nr; // number of readers
	private Semaphore sem_readers; // synchronizing access to nr
	private Semaphore sem_rw; // synchronizing Read/Write access to an object
	transient private String readLockHolder,writeLockHolder,updateLockHolder; // holders of the respective locks
	transient private String method; // the name of the method on which a lock was obtained

	public Lock() 
	{
		nr = 0;
		sem_readers = new Semaphore();
		sem_rw = new Semaphore(); 
		readLockHolder = "#";
		writeLockHolder = "#";
		updateLockHolder = "#";
		method = "#";
		
	}
	 
	
	/**
	 * Get read lock. 
	 * 
	 */
	public void getReadLock(String holder, String name)
	{
		sem_readers.P();
			nr = nr + 1;
			if(nr == 1)
			  sem_rw.P();
			readLockHolder = holder;
			method = name;
		try {
			sem_readers.V();
		} catch (SemaphoreException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void releaseReadLock()
	{
		try {
			sem_readers.P();
				nr = nr - 1;

				if(nr == 0)
				{
					readLockHolder = "";
					method = "";
					sem_rw.V();
				}				
		
		    sem_readers.V();
		    
		} catch (SemaphoreException e) {
					// TODO Auto-generated catch block
			e.printStackTrace();
		}
	
	}
	
	public void getWriteLock(String holder, String name)
	{
			  sem_rw.P();
			  writeLockHolder = holder;
			  method = name;
	}
	

	public void releaseWriteLock()
	{
			  writeLockHolder = "";
			  method = "";
			  
			  try {
					sem_rw.V();
				} catch (SemaphoreException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
	}
	
	public void getUpdateLock(String holder,String name)
	{
		sem_rw.P();
			  updateLockHolder = holder;
			  method = name;
	}
	
	public void releaseUpdateLock()
	{
			  updateLockHolder = "";
			  method = "";
			  
			  try {
					sem_rw.V();
				} catch (SemaphoreException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
	}
	
	public String getReadLockHolder()
	{
		return readLockHolder;
	}
	
	public String getWriteLockHolder()
	{
		return writeLockHolder ;
	}
	
	public String getUpdateLockHolder()
	{
		return updateLockHolder;
	}
	
	public String getMethodName()
	{
		return method;
	}
	
}
